/*************************************************************************
 * All portions of code are copyright by their respective author/s.
 * Copyright (C) 2007      Bryan Christ <bryan.christ@hp.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <pty.h>
#include <utmp.h>
#include <stropts.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <wait.h>
#include <poll.h>
#include <ctype.h>

#include <viper.h>

#include "vwmterm.h"
#include "watchdog.h"
#include "events.h"
#include "render.h"

G_MODULE_EXPORT gchar* g_module_check_init(GModule *module)
{
	gchar			*libfilename;

	/*	preload libgtop for use with this module	*/	
	if(g_module_open("libutil.so",G_MODULE_BIND_LAZY)==NULL)
 		return "\ncould not preload libutil.so";
 	
	/*	register with vwm 	*/
	libfilename=(gchar*)g_module_name(module);	
	vwm_module_add("Tools","Terminal Shell",ptyterm_init,NULL,libfilename);

	return NULL;
}

WINDOW* ptyterm_init(gpointer anything)
{
   extern WINDOW  *SCREEN_WINDOW;
	WINDOW	      *window;
	gint		      width,height;
	gchar		      *termvar;
   gchar          *title;
	PTY_TERM	      *pty_term;
   PTY_TERM       pty_clone;
	gint		      retval;

   getmaxyx(SCREEN_WINDOW,height,width);
   if(height>30 && width>84)
   {
      height=25;
      width=80;
   }
   else
   {
      /* calculate scaled window size */
	   window_get_size_scaled(NULL,&width,&height,0.85,0.65);
	   if(width>80) width=80;
	   if(height>25) height=25;
   }
		
	/* create pseudo-terminal */
	pty_term=(PTY_TERM*)g_malloc0(sizeof(PTY_TERM));
	pty_term->win_sz.ws_row=height;
	pty_term->win_sz.ws_col=width;
	tcgetattr(STDIN_FILENO,&pty_term->term_desc);

	/*	terminal settings - added 2/6/2006	*/
	/* updated for better compat on 5/2/2007	*/
	pty_term->term_desc.c_lflag |= ISIG;
	pty_term->term_desc.c_lflag |= ICANON;
	pty_term->term_desc.c_lflag |= IEXTEN;
	pty_term->term_desc.c_lflag |= ECHO;
	pty_term->term_desc.c_iflag |= BRKINT;
	pty_term->term_desc.c_iflag |= IGNPAR;
	pty_term->term_desc.c_iflag |= IXON;
	pty_term->term_desc.c_iflag &= ~IXANY;
 	pty_term->term_desc.c_iflag |= ICRNL;	/*	this has to be set for su & ssh	*/
	pty_term->term_desc.c_iflag &= ~INLCR;
	pty_term->term_desc.c_iflag &= ~IGNCR;
	pty_term->term_desc.c_oflag &= ~ONLCR;  /* suppress NL to CR-NL combo	*/
	
   /* some process will change the terminal settings and not change them
      back.  a copy of the defaults settings will be stored in order to
      reset the terminal later.  */
   memcpy(&pty_term->term_default,&pty_term->term_desc,sizeof(struct termios));
   
	retval=openpty(&pty_term->master_fd,&pty_term->slave_fd,NULL,
		&pty_term->term_desc,&pty_term->win_sz);
	if(retval==-1) return NULL;
	sprintf(pty_term->term_name," Terminal Shell: %s ",
      ttyname(pty_term->slave_fd));

   /* copy the contents of pty_term (which is on the heap) into pty_clone
      which is on the stack.  this ensures that the child process will have
      correct references to file descriptors, etc. because the stack gets
      cloned during a fork(). */
   memcpy(&pty_clone,pty_term,sizeof(PTY_TERM));
   pty_clone.child_pid=fork();
   if(pty_clone.child_pid==0)
	{
      close(pty_clone.master_fd);
      if(login_tty(pty_clone.slave_fd)==-1) return NULL;

#ifdef __sun
		ioctl(pty_clone.slave_fd,I_PUSH,"ttcompat");
#else
		ioctl(pty_clone.slave_fd,I_PUSH,"ptem");
		ioctl(pty_clone.slave_fd,I_PUSH,"ldterm");
#endif 

		signal(SIGINT,SIG_DFL);
      dup2(pty_clone.slave_fd,STDERR_FILENO);
      
		termvar=getenv("SHELL");
		if(termvar==NULL) termvar="/bin/sh";
      title=g_strrstr(termvar,"/");
      if(title!=NULL) title++;
      if(title[0]=='\0') title="sh";
      setenv("TERM","vwmterm",1);
      setenv("COLORTERM","vwmterm",1);
      setenv("LS_OPTIONS"," --color=always",1);
      execl(termvar,title,"-i",(char*)NULL);
      close(pty_clone.slave_fd);
		exit(0);
	}

   /* we used pty_clone for the fork so we need to make sure that pty_term
      has the correct child PID reference.   */   
   pty_term->child_pid=pty_clone.child_pid;

   /* create window  */
   viper_thread_enter();	
	window=viper_window_create(pty_term->term_name,0.5,0.5,width,height,TRUE);
   viper_window_set_state(window,STATE_UNSET | STATE_NORESIZE);
	viper_window_set_limits(window,15,2,WSIZE_UNCHANGED,WSIZE_UNCHANGED);
	scrollok(window,FALSE);
	pty_term->fg=COLOR_WHITE;
	pty_term->bg=COLOR_BLACK;
	wattrset(window,VIPER_COLORS(pty_term->fg,pty_term->bg));
	wbkgdset(window,VIPER_COLORS(pty_term->fg,pty_term->bg));
	werase(window);
	pty_term->scrollback=dupwin(window);
	scrollok(pty_term->scrollback,TRUE);
   pty_term->new_state |= PTY_STATE_SCROLLOK;
   pty_term->active_wnd=pty_term->scrollback;
	viper_thread_leave();

   /* initialize these so that VWMTERM_IS_FG and VWMTERM_WAS_FG macros
      will both return true.  */
	pty_term->new_state |= PTY_STATE_FG;
	pty_term->old_state |= PTY_STATE_FG;

	psthread_add("ui",ptyterm_psthread,(gpointer)window,(gpointer)pty_term);
	
	viper_thread_enter();
	viper_event_set(window,"window-resized",ptyterm_ON_RESIZE,NULL);
	viper_event_set(window,"window-close",ptyterm_ON_CLOSE,(gpointer)pty_term);
	viper_event_set(window,"window-destroy",ptyterm_ON_DESTROY,
		(gpointer)pty_term);
	viper_window_set_key_func(window,ptyterm_ON_KEYSTROKE);
	viper_window_set_userptr(window,(gpointer)pty_term);
	viper_thread_leave();
	
	return window;
}

gint ptyterm_psthread(gpointer arg1,gpointer arg2)
{
	static gchar	*acs="`afgjklmnopqrstuvwxyz{|}~";
	WINDOW  			*window;
	PTY_TERM			*pty_term;
	gint				retval=0;
	pid_t				child_pid=0;
	gchar				*buffer=NULL;
   gchar          *carryover;
	gchar				*pos_ptr;
   gint           bytes_waiting=0;
	gint				bytes_to_read=0;
	gint			   bytes_read=0;
	gint				bytes_xlate=0;
	gint				x,y;
	struct pollfd	data_poll;
	chtype			cell;
	gchar				*token;
		
	window=(WINDOW*)arg1;
	pty_term=(PTY_TERM*)arg2;
	
	/*	if child_pid==pty_term->child_pid then it is very likely that the shell
		(ie. bash) terminated normally.  in this case we can skip the
      "window-close" event and proceed to window destruction.  alternately,
      when when WHNOHANG is set, -1 indicates an error.  when this happens, it
		probably means that the user closed the window using the VWM
		window close button and the event "window-close" has already been
   	handled. */
	child_pid=waitpid(pty_term->child_pid,NULL,WNOHANG);
	if(child_pid==pty_term->child_pid || child_pid==-1)
	{
		viper_thread_enter();
		viper_window_destroy(window);
		viper_thread_leave();
		return PSTHREAD_TERMINATE;
	}

   data_poll.fd=pty_term->master_fd;
	data_poll.events=POLLIN;
   retval=poll(&data_poll,1,0);
	
	if(retval>0)
	{
 		/*	find out how much data is availbe for reading and allocate
			a buffer sufficent to store the data.  one of these _should_
         be defined on most *nix systtems.   */
#ifdef FIONREAD
		ioctl(pty_term->master_fd,FIONREAD,&bytes_waiting);
#else 
      ioctl(pty_term->master_fd,TIOCOUTQ,&bytes_waiting);
#endif
 
      bytes_to_read=bytes_waiting;
      if(bytes_to_read>PTY_READ_MAX) bytes_to_read=PTY_READ_MAX;
 		buffer=(gchar*)g_malloc0(bytes_to_read+10); /* 10 byte padding */
		pos_ptr=buffer;
		bytes_read=read(pty_term->master_fd,buffer,bytes_to_read);
		if(bytes_read==0)
      {
         g_free(buffer);
         return PSTHREAD_CONTINUE;
      }

      /* if read() returns -1, something bad has happened.  most likely the
         shell died (for some reason, maybe user sent SIGKILL).  if this
         happens we need to clean up gracefully.   */
      if(bytes_read==-1)
      {
         g_free(buffer);
         viper_thread_enter();
         viper_window_close(window);
         viper_thread_leave();
         /* allow this pseudo thread to run one more time so that waitpid()
            will catch the exit on next iteration.   */
         return PSTHREAD_CONTINUE;
      }

      /* prepend any broken escape sequences which might have been carried
         over from previous read */
      if(pty_term->carryover!=NULL)
      {
         carryover=(gchar*)g_malloc0(pty_term->carry_sz+bytes_to_read+10);
         memcpy(carryover,pty_term->carryover,pty_term->carry_sz);
         memcpy(&carryover[pty_term->carry_sz],buffer,bytes_to_read);
         g_free(buffer);
         buffer=carryover;
         g_free(pty_term->carryover);
         pty_term->carryover=NULL;
         bytes_read=bytes_to_read+pty_term->carry_sz;
         pos_ptr=buffer;
      }

		viper_thread_enter();
		while(bytes_read>0)
		{
         /* if bytes_to_read==PTY_READ_MAX, it almost certainly means ouput
            from the slave process was broken up in chunks.  this can cause
            broken escape sequences which won't get parsed correctly.  they
            need to be treated as carryover for the next read.  the saftey
            margin of 20 bytes is arbitrarily chosen.  i've never seen an
            escape code sequence longer than 20 bytes--but it could happen. */
         if((bytes_waiting>bytes_to_read) && (bytes_read<20))
         {
            ptyterm_carryover(pty_term,pos_ptr,bytes_read);
            break;
         }

			/*	discard useless escape codes	*/
			bytes_xlate=ptyterm_xlate_discards(pos_ptr);
			if(bytes_xlate>0)
			{
				bytes_read-=bytes_xlate;
				pos_ptr+=bytes_xlate;
				continue;
			}

			/* these are special 1-byte control codes.   */
			bytes_xlate=ptyterm_xlate_specials(pty_term,pty_term->active_wnd,
				pos_ptr);
			if(bytes_xlate>0)
			{
				pos_ptr+=bytes_xlate;
 				bytes_read--;
				continue;
			}
         if(bytes_xlate==-1)
         {
            ptyterm_carryover(pty_term,pos_ptr,bytes_read);
            break;
         }
		
			bytes_xlate=ptyterm_xlate_esc(pty_term,pty_term->active_wnd,pos_ptr);
         /* a return value of -1 indicates a broken escape sequence.  this
            is a timing problem which occurs with the read/write master/slave
            pty buffers fill up.  the solution is to store the incomplete
            sequence and prepend on the next read. */
         if(bytes_xlate==-1)
         {
            ptyterm_carryover(pty_term,pos_ptr,bytes_read);
            break;
         }
         else pty_term->carryover=NULL;

			if(bytes_xlate==0)
			{
				cell=pos_ptr[0];
				if(pty_term->new_state & PTY_STATE_ACS)
				{
					token=strchr(acs,pos_ptr[0]);
					if(token!=NULL) cell=NCURSES_ACS(pos_ptr[0]);
				}

            ptyterm_render_addch(pty_term,cell);
            /* vwmterm_watchdog(pty_term); */
				pos_ptr++;
				bytes_read--;
			}
			else
			{
				pos_ptr+=bytes_xlate;
				bytes_read-=bytes_xlate;
			}
		}
		
		if(buffer!=NULL) g_free(buffer);
		getmaxyx(window,y,x);
		/* it's not well documented, but getmaxyx() does not actually return the
			maximum value for x and y coords, but rather the width and height of
			a window.  the former is zero indexed while the latter is index+1.
			for this reason it is necessary to subtract 1 from y and x.	*/
		copywin(pty_term->active_wnd,window,0,0,0,0,y-1,x-1,FALSE);

		/*	draw the cursor if the cursor_invisble flag is not set	*/
      if(!(pty_term->new_state & PTY_STATE_CIVIS))
      {
         getyx(pty_term->active_wnd,y,x);
			mvwchgat(window,y,x,1,A_REVERSE,
				viper_color_pair(pty_term->fg,pty_term->bg),NULL);
      }

		/*	obligatory redraw and unlock	*/
		viper_window_redraw(window);
		viper_thread_leave();
	}
	
   /* if poll() returns -1, something bad has happened.  most likely the shell
      died (for some reason, maybe user sent SIGKILL).  if this happens we
      we need to clean up gracefully   */
	if(retval==-1)
   {
      viper_thread_enter();
      viper_window_close(window);
      viper_thread_leave();
      /* allow this pseudo thread to run one more time so that waitpid() will
         catch the exit on next iteration.   */
      return PSTHREAD_CONTINUE;
   }

	if(retval>0) return PSTHREAD_GROW;
	return PSTHREAD_CONTINUE;
}


void ptyterm_carryover(PTY_TERM *pty_term,gchar *pos_ptr,gint remainder)
{
   pty_term->carryover=(gchar*)g_malloc0(remainder);
   memcpy(pty_term->carryover,pos_ptr,remainder);
   pty_term->carry_sz=remainder;

   return;
}
   

gint ptyterm_xlate_esc(PTY_TERM *pty_term,WINDOW *window,gchar *buffer)
{
	gchar		      codes[]={'l','d','h','m','r','H','K','J','L','C'};
	gint		      array_size=0;
	gint		      size=0;
	gint		      idx;
	gint		      max_x,max_y;
	gint		      cur_x,cur_y;
	gint		      x=0,y=0;
	gint		      top,bottom;
	gint		      value1;
	gchar		      *token;
	gchar		      *control=NULL;
 						
	getmaxyx(window,max_y,max_x);
	getyx(window,cur_y,cur_x);

	/*	if its not an escape code, bail	*/
	if(buffer[0]!=27) return 0;

	/* incomplete buffer transmission	*/
	if(buffer[0]==27 && buffer[1]=='\0') return -1;

   /* handle a terminal reset (RIS) */
   if(buffer[0]==27 && buffer[1]=='c')
   {
      /* flush the master and slave pty fds when the terminal is reset.
         not sure if this is necessary, but it doesn't seem to hurt.  */
      tcdrain(pty_term->slave_fd);
      tcdrain(pty_term->master_fd);
      return 2;
   }   
   
	/* it's an escape code, but not one defined by vwmterm	*/
	if(buffer[0]==27 && buffer[1]!='[') return 0;

	buffer+=2;

	array_size=sizeof(codes)/sizeof(codes[0]);
	for(idx=0;idx<array_size;idx++)
	{
		token=strchr(buffer,codes[idx]);
		if(token!=NULL)
		{
			if(control==NULL) control=token;
			if(token<control) control=token;
		}
	}
   /* legitimate escape code suffixes are defined in the terminfo file
      for vwmterm.  if we do not find one of these suffixes, then our async
      buffer was cut during the middle of an escape transmission and so we need
      to carryover the remainder. */
	if(control==NULL) return -1;
	size=((control-buffer)/sizeof(gchar))+3;		

	if(*control=='h')
	{
      /* turn off auto margins capability (rmam)   */
		if(memcmp(buffer,"?7h",3)==0)
      {
         scrollok(window,FALSE);
         pty_term->new_state &= ~(PTY_STATE_SCROLLOK);
      }
      /* make cursor visible (cnorm)   */
      if(memcmp(buffer,"?25h",4)==0) pty_term->new_state &= (~PTY_STATE_CIVIS);
		/* start cursor contorl mode (ncurses initscr) (smcup)   */
		if(memcmp(buffer,"?1049h",6)==0)
		{
 			wattron(window,A_NORMAL);
			werase(window);
         ptyterm_render_move(pty_term,0,0);
		}
		return size;
	}

	if(*control=='d')
	{
		value1=atoi(buffer)-1;
      ptyterm_render_move(pty_term,value1,cur_x);
		return size;
	}

	if(*control=='l')
	{
      /* turn off auto margins capability (smam)   */
 		if(memcmp(buffer,"?7l",3)==0)
      {
         scrollok(window,TRUE);
         pty_term->new_state |= PTY_STATE_SCROLLOK;
      }
      /* make cursor invisible (civis) */
      if(memcmp(buffer,"?25l",4)==0) pty_term->new_state |= PTY_STATE_CIVIS;
		/* end cursor contorl mode (ncurses endwin) (rmcup)   */
		if(memcmp(buffer,"?1049l",6)==0)
		{
			ptyterm_xlate_sgr(pty_term,window,0);
			scrollok(window,TRUE);
         pty_term->new_state |= PTY_STATE_SCROLLOK;
         pty_term->new_state &= (~PTY_STATE_CIVIS);
		}
		return size;
	}

   /* set window scrolling region   */
	if(*control=='r')
	{
		top=atoi(buffer)-1;
		token=strchr(buffer,';');
		bottom=atoi(buffer)-1;
		wsetscrreg(window,top,bottom);
		return size;
	}
	
	if(*control=='m')
	{
		if(buffer==control) ptyterm_xlate_sgr(pty_term,window,0);
		while(buffer!=control)
		{
			value1=atoi(buffer);
			ptyterm_xlate_sgr(pty_term,window,value1);
			token=strchr(buffer,';');
			if(token==NULL) break;
			if(token<control) buffer=token;
			if(token>control) break;
			buffer++;
		}
		return size;
	}

	if(*control=='H')
	{
		if(buffer==control)
		{
         ptyterm_render_move(pty_term,0,0);
			return size;
		}

		y=atoi(buffer)-1;
		token=(strchr(buffer,';'));
		if(token==NULL) return size;
		buffer=token;
		buffer++;
		x=atoi(buffer)-1;
		if(x<0) x=0;
		if(y<0) y=0;
      ptyterm_render_move(pty_term,y,x);
		return size;
	}
		
	if(*control=='J')
	{
      if(memcmp(buffer,"2J",2)==0) ptyterm_render_move(pty_term,0,0);
      wclrtobot(window);
		return size;
	}

	if(*control=='K')
	{
		if(buffer[0]=='K')
		{
         wclrtoeol(window);
         /* ptyterm_render_clrtoeol(pty_term); */
         wchgat(window,-1,A_NORMAL,viper_color_pair(pty_term->fg,pty_term->bg),
            NULL);
		}
		return size;
	}

	if(*control=='L')
	{
		if(buffer[0]=='L') winsertln(window);
		return size;
	}

   if(*control=='C')
   {
      /* this should never happend  */
      if(buffer==control) return size;
   
      x=atoi(buffer);
	   if(x<0) x=0;
      /* wmove(window,cur_y,cur_x+x);*/
      ptyterm_render_move(pty_term,cur_y,cur_x+x);
      return size;
   }

	return -1;
}

gint	ptyterm_xlate_specials(PTY_TERM *pty_term,WINDOW *window,gchar *buffer)
{
	gint           cur_x,cur_y;
	gint	         max_x,max_y;
	guint	         code;
 
	code=(guint)buffer[0];
	if(code>31 && code <127) return 0;

	getmaxyx(window,max_y,max_x);
	getyx(window,cur_y,cur_x);
	
	switch(code)
	{
		case 0: return 0;
		case 3:
		{
			scrollok(window,TRUE);
         pty_term->new_state |= PTY_STATE_SCROLLOK;
			return 1;
		}
		case 7:
		{
			beep();
			return 1;
		}
		case 8:
		{	
         ptyterm_render_move(pty_term,cur_y,cur_x-1);
			return 1;
		}
		case 9:
      {
         /* the best way to handle a TAB is to simply pass it through
            unhandled and write it to the window.  this strategy relies on
            ncurses to do all of the tab stop calculation.  */
         return 0;
      }
		case 10: /* NL or LF (new line / line feed)  */
		{
         ptyterm_render_addch(pty_term,'\n');
			return 1; 
		}
		case 13: /* CR (carriage return) */
		{	
         /* a very common terminal setting is to translate the NL (linefeed/LF)
            to 2 bytes (CR+NL) and this causes problems rendering some
            applications.  this setting is initially suppressed when the
            terminal is initialized, but ultimately we have no control over
            what an application sets with tcsetattr().  if this sequence is
            detected we need to skip the CR and just process the NL. */
         if(buffer[1]=='\n') return 1;

         ptyterm_render_addch(pty_term,'\r');
			return 1;
		}
		case 14:
		{
			/* wmove(window,cur_y+1,cur_x); */
         ptyterm_render_move(pty_term,cur_y+1,cur_x);
			return 1;
		}
		case 15: return 1;
		case 16:
		{
			/* wmove(window,cur_y-1,cur_x); */
         ptyterm_render_move(pty_term,cur_y-1,cur_x);
			return 1;
		}
      /* don't process 27 because it is the escape code for more complex
         control sequences.   */
		case 27: return 0;
      /* ignore odd escape codes that aren't currently handled.   */
		default: return 1;
	}

	return 1;
}

void ptyterm_xlate_sgr(PTY_TERM *pty_term,WINDOW *window,gint code)
{
	gint		colors[]={COLOR_BLACK,COLOR_RED,COLOR_GREEN,COLOR_YELLOW,
					COLOR_BLUE,COLOR_MAGENTA,COLOR_CYAN,COLOR_WHITE};
	
	if(code==0) 
	{
		wattrset(window,A_NORMAL);
		pty_term->fg=COLOR_WHITE;
		pty_term->bg=COLOR_BLACK;
		wattron(window,VIPER_COLORS(COLOR_WHITE,COLOR_BLACK));
		return;
	}

	if(code==1) wattron(window,A_BOLD);       /* bold  */
	if(code==2) wattron(window,A_DIM);        /* dim   */
	if(code==3) wattron(window,A_STANDOUT);   /* standout */
	if(code==4) wattron(window,A_UNDERLINE);  /* underline   */ 
	if(code==5) wattron(window,A_BLINK);      /* blink */
	if(code==6) wattron(window,A_BLINK);      /* blink */
	if(code==7) wattron(window,A_REVERSE);    /* reverse  */
	if(code==8) wattron(window,A_INVIS);      /* invisible   */
	/*	codes 10 and 11 are reserved in the ansi spec so they get used
		for entrance and exit escapes for ALT CHARSET.  this is defined in
		the terminfo for vwmterm.  everybody else is doing it so why
		can't i? :)	*/
	if(code==10) pty_term->new_state &= (~PTY_STATE_ACS);
	if(code==11) pty_term->new_state |= PTY_STATE_ACS;
	if(code==23) wattroff(window,A_STANDOUT); 
	if(code==24) wattroff(window,A_UNDERLINE);
	if(code==25) wattroff(window,A_BLINK);
	if(code==27) wattroff(window,A_REVERSE);
	if(code>=30 && code<=37) pty_term->fg=colors[code-30];
	if(code==39) pty_term->fg=COLOR_WHITE;	            /* hard-coded default	*/ 
	if(code>=40 && code<=47) pty_term->bg=colors[code-40];
	if(code==49) pty_term->bg=COLOR_BLACK;	            /* hard-coded default   */

	if(code>=30 && code<=49) 
		wattron(window,VIPER_COLORS(pty_term->fg,pty_term->bg));

	return;
}

/*	some programs try to second-guess ncurses.  this is really unfortunate
	because the programs can generate escape codes that are unhandled.  the
	effects of these codes on the parser are generally not good.  this
	function attempts to discard these bad values.	*/
gint ptyterm_xlate_discards(gchar *buffer)
{
	/* this list requires a NULL terminator	*/

	gchar	*discards[]={
				"\x1B[?1000h","\x1B)0","\x1B[?1000l","\x1B[?9h","\x1B[?9l",
				"\x1B\x38","\x1B\x37"};
   static guint   array_sz=sizeof(discards)/sizeof(discards[0]);
	gint	         size=0;
	guint	         i=0;


   for(i=0;i<array_sz;i++)
   {
      size=strlen(discards[i]);  
      if(memcmp(discards[i],buffer,size)==0) break;
   }

   if(i==array_sz) return 0;
	return size;
}
		


	
				
